#!/usr/bin/env python

from .src.data_cleaning import DataCleaner
from .src.modelzoo import ModelPipeLine
from .src.ensembling import Ensembling
from .src.economy import UnitEconomy
from .src.data_cleaning import ReduceVIF


import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout,BatchNormalization
from tensorflow.keras.optimizers import Adam, Nadam, AdamW
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.metrics import Recall, Precision
from tensorflow.keras.regularizers import l2
from tensorflow.keras.callbacks import ReduceLROnPlateau
from tensorflow.keras.optimizers import SGD

import numpy as np
import pandas as pd
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
import plotly.express as px
import time
import re

# Pre-Processing
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder, MinMaxScaler
from sklearn.model_selection import train_test_split, GridSearchCV, cross_val_score, cross_validate
from sklearn.impute import KNNImputer
from sklearn.neighbors import LocalOutlierFactor
from scipy import stats
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.inspection import permutation_importance
from sklearn.utils.fixes import parse_version
from imblearn.over_sampling import SMOTE
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import CategoricalNB
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier, VotingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import roc_auc_score

from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score


# Metrics
from sklearn.metrics import multilabel_confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import roc_auc_score, roc_curve, precision_recall_curve, average_precision_score
from sklearn.metrics import RocCurveDisplay

from abc import ABC, abstractmethod

class Ensembling:

    def __init__(self, model_zoo, base_models, model_list=["LogRegression"]):
        self.model_list = model_list
        self.base_models = base_models
        self.model_zoo = model_zoo

    def runEnsemble(self,performance,shape):
        result = []

        Y_preds = []
        for model_name in self.model_list:
            model = self.base_models[model_name](performance=performance,shape=shape)

            start = time.time()
            model = self.model_zoo.model_fit(model)
            Y_pred, Y_pred_classes = self.model_zoo.model_predict(model)
            end = time.time()

            Y_preds.append(Y_pred)
            result.append({"Model": model_name,
                           "Class_probabilities": Y_pred,
                           "Classes": Y_pred_classes,
                           "Elapsed_Time": end - start,
                           })

        # Averaging results
        Y_pred = np.mean(Y_preds, axis=0)
        Y_pred_df = pd.DataFrame(Y_pred)
        Y_pred_df["grade"] = Y_pred_df.idxmax(axis=1)
        Y_pred_classes = Y_pred_df["grade"]

        return Y_pred, Y_pred_classes, result



import numpy as np
import pandas as pd
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
import plotly.express as px
import time
import re

# Pre-Processing
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder, MinMaxScaler
from sklearn.model_selection import train_test_split, GridSearchCV, cross_val_score, cross_validate
from sklearn.impute import KNNImputer
from sklearn.neighbors import LocalOutlierFactor
from scipy import stats
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.inspection import permutation_importance
from sklearn.utils.fixes import parse_version
from imblearn.over_sampling import SMOTE
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import CategoricalNB
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier, VotingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import roc_auc_score

from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score

# Metrics
from sklearn.metrics import multilabel_confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import roc_auc_score, roc_curve, precision_recall_curve, average_precision_score
from sklearn.metrics import RocCurveDisplay

class DataCleaner():

    def __init__(self, df, target_column):

        self.df = df
        self.target_column = target_column

    def select_train(self):
        train_columns = self.df.columns.tolist()
        train_columns = [x for x in train_columns if x != self.target_column]

        return train_columns

    def clean_non_numerical(self, columns=[]):
        for column in columns:
            # Remove non-numeric characters
            self.df[column] = self.df[column].apply(lambda x: re.sub(r'[^0-9.]', '', str(x)))
            # Replace hex values with 10x numbers
            self.df[column] = self.df[column].apply(
                lambda x: int(x, 16) if isinstance(x, str) and x.startswith('0x') else x)
            # Replace empty strings with NaN
            self.df[column] = self.df[column].replace('', np.nan)
            # Convert dtype to float
            self.df[column] = self.df[column].astype(float)
            # Convert negatives like '-1' to NaN
            # df[df[column]<0] = df[df[column]<0].replace('', np.nan)
            # Convert dtype to float and coerce errors to NaN (invalid strings to NaN)
            self.df[column] = pd.to_numeric(self.df[column], errors='coerce')

        return self.df

    def categorical_to_onehot(self, columns=[]):
        for column in columns:
            unique_types = self.df[column].dropna().str.split(',').explode().str.strip().unique()

            # For each of value types:
            for i in self.df[column].value_counts().head(9).index[1:]:
                # Create a new column
                # Set value = 1 if the value type is present in the initial cell value, 0 otherwise
                self.df[column + "_" + i] = self.df[column].str.contains(i, na=False).astype(float)

            # Delete the original column after creating binary columns
            del self.df[column]
        return self.df

    def exists_or_not(self, columns=[]):
        for column in columns:
            # Replace empty strings with 0
            self.df[column] = self.df[column].replace('', 0)
            # Replace all other values with 1
            self.df.loc[self.df[column] != 0, column] = 1.0
            self.df[column] = self.df[column].astype(float)

        return self.df

    def select_missing_values(self, dataframe):
        selected_columns = dataframe.columns[dataframe.isna().any()].tolist()

        return selected_columns

    def knn_impute(self, n=5):

        # Apply KNN imputation to the specified column
        imputer = KNNImputer(missing_values=np.nan, n_neighbors=n)

        train_columns = self.select_train()
        selected_columns = self.select_missing_values(self.df[train_columns])

        self.df[selected_columns] = imputer.fit_transform(self.df[selected_columns])

        return self.df

    def remove_outliers(self, threshold):

        train_columns = self.select_train()

        ids = []
        for column in train_columns:
            z = np.abs(stats.zscore(self.df[column]))
            outlier_ids = np.where(z > threshold)[0]

            ids.extend(outlier_ids)

        ids = list(set(ids))
        self.df = self.df.drop(ids)

        return self.df


class ReduceVIF(BaseEstimator, TransformerMixin):
    def __init__(self, thresh=10):
        # Values between 5 and 10 can be considered acceptable.
        # Values above 10 are to be removed.
        self.thresh = thresh

    def fit(self, X, y=None):
        print('ReduceVIF fit')
        return self

    def transform(self, X, y=None):
        print('ReduceVIF transform')
        columns = X.columns.tolist()
        return ReduceVIF.calculate_vif(X, self.thresh)

    @staticmethod
    def calculate_vif(X, thresh=5.0):
        # Taken from https://stats.stackexchange.com/a/253620/53565 and modified
        dropped = True
        while dropped:
            variables = X.columns
            dropped = False
            vif = [variance_inflation_factor(X[variables].values, X.columns.get_loc(var)) for var in X.columns]

            max_vif = max(vif)
            if max_vif > thresh:
                maxloc = vif.index(max_vif)
                print(f'Dropping {X.columns[maxloc]} with vif={max_vif}')
                X = X.drop([X.columns.tolist()[maxloc]], axis=1)
                dropped = True
        return X



import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout,BatchNormalization
from tensorflow.keras.optimizers import Adam, Nadam, AdamW
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.metrics import Recall, Precision
from tensorflow.keras.regularizers import l2
from tensorflow.keras.callbacks import ReduceLROnPlateau
from tensorflow.keras.optimizers import SGD

import numpy as np
import pandas as pd
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
import plotly.express as px
import time
import re

# Pre-Processing
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder, MinMaxScaler
from sklearn.model_selection import train_test_split, GridSearchCV, cross_val_score, cross_validate
from sklearn.impute import KNNImputer
from sklearn.neighbors import LocalOutlierFactor
from scipy import stats
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.inspection import permutation_importance
from sklearn.utils.fixes import parse_version
from imblearn.over_sampling import SMOTE
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import CategoricalNB
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier, VotingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import roc_auc_score

from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score


# Metrics
from sklearn.metrics import multilabel_confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import roc_auc_score, roc_curve, precision_recall_curve, average_precision_score
from sklearn.metrics import RocCurveDisplay

from abc import ABC, abstractmethod


class UnitEconomy:
    """
    Clients with Poor credit score are rejected (i.e. no loan is issued).
    Depending on the correctness of risk evaluation each approved client results in the given profit or loss in arbitrary currency units (ACU):

    - Poor client evaluated as Standard yields loss of ACU 200.
    - Poor client evaluated as Good yields loss of ACU 400.

    - Standard client evaluated as Standard yields profit of ACU 50.
    - Standard client evaluated as Good yields profit of ACU 30.

    - Good client evaluated as Standard yields profit of ACU 70.
    - Good client evaluated as Good yields profit of ACU 100.

    - Each client evaluation regardless of outcome costs ACU 5.
    """

    def __init__(self, **kwargs):
        self.eval_cost = kwargs["eval_cost"]
        self.poor_as_standard = kwargs["poor_as_standard"]
        self.poor_as_good = kwargs["poor_as_good"]
        self.standard_as_standard = kwargs["standard_as_standard"]
        self.standard_as_good = kwargs["standard_as_good"]
        self.good_as_standard = kwargs["good_as_standard"]
        self.good_as_good = kwargs["good_as_good"]

        """
        COST-BENEFIT MATRIX:

        Actual type | Evaluated as Poor | Evaluated as Standard | Evaluated as Good
        ___________________________________________________________________________

        Poor        |        -5         |       -5-200          |      -5-400
        ___________________________________________________________________________

        Standard    |        -5         |       -5+50           |      -5+30
        ___________________________________________________________________________

        Good        |        -5         |       -5+70           |      -5+100
        ___________________________________________________________________________        
        """

        self.cost_benefit_matrix = np.array(
            [[-self.eval_cost, -self.eval_cost + self.poor_as_standard, -self.eval_cost + self.poor_as_good],
             [-self.eval_cost, -self.eval_cost + self.standard_as_standard, -self.eval_cost + self.standard_as_good],
             [-self.eval_cost, -self.eval_cost + self.good_as_standard, -self.eval_cost + self.good_as_good]])

    # NumPy version of custom loss function
    def custom_loss_np(self, Y_true, Y_pred):
        """
        Custom loss function to compute the profit or loss based on the cost-benefit matrix using NumPy.

        Args:
        Y_true: true class labels as a NumPy array (0 = Poor, 1 = Standard, 2 = Good)
        Y_pred: predicted class probabilities as a NumPy array (shape: [n_samples, n_classes])

        Returns:
        A scalar representing the average loss across all examples.
        """

        # Ensure Y_true is an integer array
        Y_true = np.squeeze(Y_true.astype(int))

        # Create one-hot encoding for true labels
        Y_true_one_hot = np.eye(len(self.cost_benefit_matrix))[Y_true]

        # Calculate expected loss
        expected_loss = np.sum(Y_pred * np.dot(Y_true_one_hot, self.cost_benefit_matrix), axis=1)

        return -np.mean(expected_loss)

    # TensorFlow version of custom loss function
    def custom_loss_tf(self, Y_true, Y_pred):
        """
        Custom loss function to compute the profit or loss based on the cost-benefit matrix.

        Args:
        Y_true: true class labels (0 = Poor, 1 = Standard, 2 = Good)
        Y_pred: predicted class probabilities or labels

        Returns:
        A scalar representing the total loss across all examples.
        """

        cost_benefit = tf.constant(self.cost_benefit_matrix, dtype=tf.float32)

        # Get the predicted class by taking the argmax (index of the highest probability)
        Y_pred_labels = tf.argmax(Y_pred, axis=1, output_type=tf.int32)

        # Ensure y_true is an integer tensor
        Y_true = tf.squeeze(tf.cast(Y_true, tf.int32))
        Y_true_one_hot = tf.one_hot(Y_true, depth=3)

        # Stack y_true and y_pred_labels into pairs: shape (batch_size, 2)
        indices = tf.stack([Y_true, Y_pred_labels], axis=1)

        expected_loss = tf.reduce_sum(Y_pred * tf.matmul(Y_true_one_hot, cost_benefit), axis=1)
        expected_loss = tf.cast(expected_loss, tf.float32)

        return -tf.reduce_mean(expected_loss)


import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout,BatchNormalization
from tensorflow.keras.optimizers import Adam, Nadam, AdamW
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.metrics import Recall, Precision
from tensorflow.keras.regularizers import l2
from tensorflow.keras.callbacks import ReduceLROnPlateau
from tensorflow.keras.optimizers import SGD

import numpy as np
import pandas as pd
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
import plotly.express as px
import time
import re

# Pre-Processing
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder, MinMaxScaler
from sklearn.model_selection import train_test_split, GridSearchCV, cross_val_score, cross_validate
from sklearn.impute import KNNImputer
from sklearn.neighbors import LocalOutlierFactor
from scipy import stats
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.inspection import permutation_importance
from sklearn.utils.fixes import parse_version
from imblearn.over_sampling import SMOTE
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import CategoricalNB
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier, VotingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelBinarizer
from sklearn.metrics import roc_auc_score

from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score


# Metrics
from sklearn.metrics import multilabel_confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import roc_auc_score, roc_curve, precision_recall_curve, average_precision_score
from sklearn.metrics import RocCurveDisplay

from abc import ABC, abstractmethod

class ModelPipeLine:

    def __init__(self, X_train, Y_train, X_test, Y_test):
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test

        self.performance = "Light"

        self.base_models = {"LogRegression": LogRegression,
                            "DecisionTree": DecisionTree,
                            "RandomForest": RandomForest,
                            "GradientBoost": GradientBoost,
                            "Bayes": Bayes,
                            "NeuralNet": NeuralNet
                            }

    def scaling(self):
        scaler = MinMaxScaler()

        X_train_scaled = scaler.fit_transform(self.X_train)
        X_test_scaled = scaler.transform(self.X_test)

        self.X_train = pd.DataFrame(X_train_scaled, columns=self.X_train.columns)
        self.X_test = pd.DataFrame(X_test_scaled, columns=self.X_test.columns)

    def oversample(self):
        # SMOTE oversampling
        smote = SMOTE(sampling_strategy='auto', random_state=42)
        self.X_train, self.Y_train = smote.fit_resample(self.X_train, self.Y_train)

    def model_fit(self, abstract_model):
        abstract_model.fit(self.X_train, self.Y_train)
        return abstract_model

    def model_predict(self, abstract_model):
        Y_pred = abstract_model.predict_proba(self.X_test)
        Y_pred_df = pd.DataFrame(Y_pred)
        Y_pred_df["grade"] = Y_pred_df.idxmax(axis=1)
        Y_pred_classes = Y_pred_df["grade"]
        return Y_pred, Y_pred_classes

    def show_results(self, Y_pred_classes):
        matrix = confusion_matrix(self.Y_test, Y_pred_classes)
        print("Accuracy: ", matrix.diagonal() / matrix.sum(axis=1))
        disp = ConfusionMatrixDisplay(matrix)
        disp.plot()

    def showROC(self, Y_pred, class_id):
        label_binarizer = LabelBinarizer().fit(self.Y_train)
        y_onehot_test = label_binarizer.transform(self.Y_test)
        y_onehot_test.shape  # (n_samples, n_classes)

        display = RocCurveDisplay.from_predictions(
            y_onehot_test[:, class_id],
            Y_pred[:, class_id],
            name=f"Class #{class_id} vs the rest",
            color="darkorange",
            plot_chance_level=True,
        )
        _ = display.ax_.set(
            xlabel="False Positive Rate",
            ylabel="True Positive Rate",
            title="One-vs-Rest ROC curve",
        )

        micro_roc_auc_ovr = roc_auc_score(
            self.Y_test,
            Y_pred,
            multi_class="ovr",
            average="micro",
        )

        print(f"Micro-averaged One-vs-Rest ROC AUC score:\n{micro_roc_auc_ovr:.2f}")


class ClassicModel(ABC):

    def __init__(self, performance, shape):
        self.set_model(performance="Light", shape=shape)
        self.performance = performance

    def fit(self, X_train, Y_train):
        print("\n Using model: ", self.model)
        self.model.fit(X_train, Y_train)
        return self.model

    def predict_proba(self, X_test):
        Y_pred = self.model.predict_proba(X_test)
        return Y_pred

    @abstractmethod
    def set_model(self) -> None:
        pass


# Model Zoo

class LogRegression(ClassicModel):
    def set_model(self, performance, shape):
        self.model = LogisticRegression(random_state=42)

    # Custom scorer function (convert it to work with make_scorer)
    def custom_scorer(estimator, X, Y):
        # Get predicted probabilities from the model
        y_pred_probs = estimator.predict_proba(X)

        # Return custom loss (as a positive value, since scorers maximize)
        return custom_loss_np(y, y_pred_probs)


class DecisionTree(ClassicModel):
    def set_model(self, performance, shape):
        self.model = tree.DecisionTreeClassifier()


class RandomForest(ClassicModel):
    def set_model(self, performance, shape):
        performance_levels = {
            "Light": {'n_estimators': 1, 'random_state': 42},
            "Medium": {'n_estimators': 50, 'random_state': 42},
            "High": {'n_estimators': 1000, 'random_state': 42}
        }
        self.model = RandomForestClassifier(**performance_levels[performance])


class GradientBoost(ClassicModel):
    def set_model(self, performance, shape):
        performance_levels = {
            "Light": {'n_estimators': 3, 'learning_rate': 1.0, 'max_depth': 3, 'random_state': 42},
            "Medium": {'n_estimators': 10, 'learning_rate': 1e-1, 'max_depth': 8, 'random_state': 42},
            "High": {'n_estimators': 20, 'learning_rate': 5e-2, 'max_depth': 10, 'random_state': 42},
        }
        self.model = GradientBoostingClassifier(**performance_levels[performance])


class Bayes(ClassicModel):
    def set_model(self, performance, shape):
        self.model = CategoricalNB()


class NeuralNet(ClassicModel):
    def set_model(self, performance, shape):
        self.performance = performance
        self.performance_levels = {
            "Light": {'epochs': 20, 'learning_rate': 1e-3},
            "Medium": {'epochs': 100, 'learning_rate': 1e-3},
            "High": {'epochs': 500, 'learning_rate': 1e-3},
        }

        model = Sequential([
            Dense(512, input_dim=shape, activation='relu'),
            BatchNormalization(),
            Dropout(0.3),
            Dense(256, activation='relu'),
            BatchNormalization(),
            Dropout(0.3),
            Dense(128, activation='relu'),
            BatchNormalization(),
            Dropout(0.25),
            Dense(64, activation='relu'),
            BatchNormalization(),
            Dropout(0.2),
            Dense(3, activation='softmax')
        ])

        try:
            Economy
        except NameError:
            loss_function = "sparse_categorical_crossentropy"
        else:
            loss_function = Economy.custom_loss_tf

        model.compile(optimizer=Adam(learning_rate=self.performance_levels[performance]["learning_rate"]),
                      # loss='sparse_categorical_crossentropy',
                      loss=loss_function,
                      metrics=['accuracy'])

        self.model = model

    def fit(self, X_train, Y_train):
        X_train_tr, X_val, Y_train_tr, Y_val = train_test_split(
            X_train, Y_train, test_size=0.1, random_state=42)

        early_stopping = EarlyStopping(monitor='val_accuracy',
                                       patience=60,
                                       mode="auto",
                                       restore_best_weights=True)

        reduce_lr = ReduceLROnPlateau(monitor='val_loss',
                                      factor=0.5,  # Learning rate reduced by half if no improvement
                                      patience=10,  # Wait 10 epochs before reducing learning rate
                                      min_lr=1e-6,  # Minimum learning rate set
                                      verbose=1)

        history = self.model.fit(
            x=X_train_tr,
            y=Y_train_tr,
            validation_data=(X_val, Y_val),
            batch_size=1024,
            epochs=self.performance_levels[self.performance]["epochs"],
            verbose=0,
            callbacks=[early_stopping, reduce_lr])

    def predict_proba(self, X_test):
        Y_pred = self.model.predict(X_test)
        return Y_pred


#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

def get_install_requirements():
    with open('requirements.txt', 'r', encoding='utf-8') as f:
        reqs = [x.strip() for x in f.read().splitlines()]
    reqs = [x for x in reqs if not x.startswith('#')]
    return reqs

setup(
    name='easycreditscoring',
    version='0.0.2',
    description='Simple library combining efficient credit scoring ML methods',
    author='miosipof',
    packages=find_packages(),
    install_requires=get_install_requirements(),
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    # test_suite='tests',
)